#include	<sys/types.h>
#include	<sys/param.h>
#include	<time.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>
#include	<errno.h>
#include	<unistd.h>

#include	<db.h>

#include	"nbs.h"


/*
	Marcus J. Ranum, 2004   - All rights reserved
	$Id: db.c,v 1.2 2004/09/03 01:36:23 mjr Exp $

	This software may be used and redistributed free of charge,
	but may not be incorporated into a commercial product or
	offering without the author's permission. Please see
	the included "LICENSE.pdf" for the software's governing
	license and terms.

	NBS - "Never Before Seen" anomaly detector database methods
	most of the actual code for nbs is done in here; you could
	replace this logic with a different database, however, if
	you care to.
*/


extern	int	errno;


static	char	dberrbuf[BUFSIZ];
static	char	*dbnotopen = "database is not open";
static	char	*internal_dbbanner = "NBS Database\n";
static	DBT	key;
static	DBT	data;
static	time_t	db_timenow = (time_t)0;



char	*
nbs_dberror()
{
	dberrbuf[BUFSIZ - 2] = '\0';
	if(dberrbuf[0] != '\0')
		return(dberrbuf);
	return("no error");
}



/* compare an update index record */
static	int
bt_cmp_upd(DB *d,const DBT *k1,const DBT *k2)
{
	NBS_UPD	u1;
	NBS_UPD	u2;


	memcpy(&u1,k1->data,sizeof(NBS_UPD));
	memcpy(&u2,k2->data,sizeof(NBS_UPD));

	if(u1.upd == u2.upd)
		return((int)(u1.off - u2.off));
	return((int)(u1.upd - u2.upd));
}



/* compare a count index record */
static	int
bt_cmp_counts(DB *d,const DBT *k1,const DBT *k2)
{
	NBS_CNT	c1;
	NBS_CNT	c2;


	memcpy(&c1,k1->data,sizeof(NBS_CNT));
	memcpy(&c2,k2->data,sizeof(NBS_CNT));

	if(c1.cnt == c2.cnt)
		return((int)(c1.off - c2.off));
	return(c1.cnt - c2.cnt);
}



/* write out a db rec file header */
static	int
db_rec_write(NBS_DB *d)
{
	NBS_HDR	b;

	strcpy(b.warn,internal_dbbanner);
	b.flg = d->flg;
	b.ecnt = d->ecnt;
	b.tlin = d->tlin;
	b.lupd = d->lupd;
	b.lopd = d->lopd;
	b.cred = d->cred;
	if(fseek(d->rec,0L,SEEK_SET) != 0L)
		return(1);
	if(fwrite(&b,sizeof(b),1,d->rec) != 1)
		return(1);
	return(0);
}




/* set up the structs in a DB header */
static	int
db_rec_setup(NBS_DB *d)
{
	NBS_HDR	b;

	b.flg = d->flg;
	b.ecnt = d->ecnt = 0;
	b.tlin = d->tlin = 0;
	time(&b.lupd);
	db_timenow = b.lupd;
	d->lupd = d->lopd = b.lopd = d->cred = b.cred = b.lupd;
	return(db_rec_write(d));
}



/* read the config from the header */
static	int
db_rec_configure(NBS_DB *d)
{
	NBS_HDR	b;

	if(fseek(d->rec,0L,SEEK_SET) != 0L)
		return(1);
	if(fread(&b,sizeof(b),1,d->rec) != 1)
		return(1);
	if(strncmp(b.warn,internal_dbbanner,strlen(internal_dbbanner) - 1)) {
		strcpy(dberrbuf,"Bad magic number in record file");
		return(1);
	}
	d->flg = b.flg;
	d->ecnt = b.ecnt;
	d->tlin = b.tlin;
	d->lupd = b.lupd;
	d->lopd = b.lopd;
	d->cred = b.cred;
	time(&db_timenow);
	return(0);
}



/* separated from the open function for clarity */
static	NBS_DB	*
internal_nbs_dbopen(char *f,int flg,int create)
{
	char	fbuf[MAXPATHLEN];
	int	x;
	int	y;
	char	*p;
	NBS_DB	*r;

	dberrbuf[0] = '\0';

	r = (NBS_DB *)malloc(sizeof(NBS_DB));
	if(r == (NBS_DB *)0) {
		/* cannot go to bailout here */
		strcpy(dberrbuf,strerror(errno));
		return((NBS_DB *)0);
	}
	r->prim = (DB *)0;
	r->cnt = (DB *)0;
	r->upd = (DB *)0;
	r->rec = (FILE *)0;

	/* open/setup record file */
	strcpy(fbuf,f);
	strcat(fbuf,".rec");
	if(create)
		p = "w";
	else
		p = "r+";
	if((r->rec = fopen(fbuf,p)) == (FILE *)0) {
		/* cannot go to bailout here */
		strcpy(dberrbuf,strerror(errno));
		free(r);
		return((NBS_DB *)0);
	}

#ifdef	NBS_DEFAULTLOCK
	if(lockf(fileno(r->rec),F_LOCK,0) != 0) {
		strcpy(dberrbuf,strerror(errno));
		fclose(r->rec);
		free(r);
		return((NBS_DB *)0);
	}
#endif

	/* read/set config from the record file */
	if(create) {
		r->flg = flg;
		x = db_rec_setup(r);
	} else
		x = db_rec_configure(r);
	if(x) {
		/* cannot go to bailout here */
		(void)fclose(r->rec);
		free(r);
		return((NBS_DB *)0);
	}


	/* open/create primary index */
	strcpy(fbuf,f);
	strcat(fbuf,".idx");
	if((x = db_create(&r->prim,NULL,0)) != 0)
		goto bailout;
	if(create)
		y = DB_CREATE|DB_EXCL;
	else
		y = 0;
	if((x = r->prim->open(r->prim,NULL,fbuf,NULL,DB_BTREE,y,0664)) != 0)
		goto bailout;


	/* open/create counts index */
	if(r->flg & NBS_DB_FLG_NOCOUNT)
		goto skipcounts;
	strcpy(fbuf,f);
	strcat(fbuf,".cnt");
	if((x = db_create(&r->cnt,NULL,0)) != 0)
		goto bailout;
	/* set comparison and prefix funcs in primary tree */
	r->cnt->set_bt_compare(r->cnt,bt_cmp_counts);
	if(create)
		y = DB_CREATE|DB_EXCL;
	else
		y = 0;
	if((x = r->cnt->open(r->cnt,NULL,fbuf,NULL,DB_BTREE,y,0664)) != 0)
		goto bailout;


skipcounts:
	/* open/create updates index */
	if(r->flg & NBS_DB_FLG_NOUPTIM)
		goto skipupdate;
	strcpy(fbuf,f);
	strcat(fbuf,".upd");
	if((x = db_create(&r->upd,NULL,0)) != 0)
		goto bailout;
	/* set comparison and prefix funcs in primary tree */
	r->upd->set_bt_compare(r->upd,bt_cmp_upd);
	if(create)
		y = DB_CREATE|DB_EXCL;
	else
		y = 0;
	if((x = r->upd->open(r->upd,NULL,fbuf,NULL,DB_BTREE,y,0664)) != 0)
		goto bailout;

skipupdate:
	return(r);

	/* notreached normally */
bailout:
	strncpy(dberrbuf,db_strerror(x),BUFSIZ - 2);
	if(r->prim != (DB *)0)
		(void)r->prim->close(r->prim,0);
	if(r->cnt != (DB *)0)
		(void)r->cnt->close(r->cnt,0);
	if(r->upd != (DB *)0)
		(void)r->upd->close(r->upd,0);
	if(r->rec != (FILE *)0)
		(void)fclose(r->rec);
	free(r);
	return((NBS_DB *)0);
}




int
nbs_dbclose(NBS_DB *d)
{
	int	x;

	dberrbuf[0] = '\0';
	if(db_rec_write(d))
		return(1);

	if((x = d->prim->close(d->prim,0)) != 0) {
		strcpy(dberrbuf,db_strerror(x));
		strcat(dberrbuf," - primary index");
		return(1);
	}


	if((d->flg & NBS_DB_FLG_NOCOUNT) == 0) {
		if((x = d->cnt->close(d->cnt,0)) != 0) {
			strcpy(dberrbuf,db_strerror(x));
			strcat(dberrbuf," - count index");
			return(1);
		}
	}


	if((d->flg & NBS_DB_FLG_NOUPTIM) == 0) {
		if((x = d->upd->close(d->upd,0)) != 0) {
			strcpy(dberrbuf,db_strerror(x));
			strcat(dberrbuf," - update index");
			return(1);
		}
	}

#ifdef	NBS_DEFAULTLOCK
	if(lockf(fileno(d->rec),F_ULOCK,0) != 0) {
		strcpy(dberrbuf,strerror(errno));
		return(1);
	}
#endif

	/* first to open, last to close! */
	if(d->rec != (FILE *)0 && fclose(d->rec) != 0) {
		strcpy(dberrbuf,strerror(errno));
		strcat(dberrbuf," - record file");
		return(1);
	}

	free(d);
	return(0);
}



NBS_DB	*
nbs_dbopen(char *f)
{
	return(internal_nbs_dbopen(f,0,0));
}



NBS_DB	*
nbs_dbcreat(char *f,int flg)
{
	return(internal_nbs_dbopen(f,flg,1));
}



/* check to see if something is in the primary index */
off_t
nbs_dbck(NBS_DB *d,char *s,int l,int *siz)
{
	int	r;
	NBS_PRI	b;

	if(d == (NBS_DB *)0 || d->prim == (DB *)0) {
		strcpy(dberrbuf,dbnotopen);
		return(-1);
	}

	memset(&key,0,sizeof(DBT));
	memset(&data,0,sizeof(DBT));
	key.data = s;
	key.size = l - 1; /* nulls not stored in btree */
	r = d->prim->get(d->prim,NULL,&key,&data,0);
	if(r == 0 && r != DB_NOTFOUND && data.size == sizeof(NBS_PRI)) {
		memcpy(&b,data.data,sizeof(b));
		if(siz != (int *)0)
			*siz = b.siz;
		return(b.off);
	}
	if(r != DB_NOTFOUND) {
                strncpy(dberrbuf,db_strerror(r),BUFSIZ - 2);
		return(-1);
	}
	return(0);
}



/* allocate new space for a new record
by the simple and cheesy method of finding
where the end of the file happens to be... */
off_t
nbs_dbnewrec(NBS_DB *d,int l,int l2,int *siz)
{
	*siz = -1;
	if(fseek(d->rec,0,SEEK_END) != 0)
		return((off_t)0);
	*siz = sizeof(NBS_REC) + l + l2;
	return(ftell(d->rec));
}



/* pull up a record from the database.
memory is retained across calls; nothing
else should ever interfere with it */
NBS_REC	*
nbs_dbgetrec(NBS_DB *d,off_t rec,int siz)
{
	static	char	*db_recbuf = (char *)0;
	static	int	db_recbufsiz = 0;
	static	NBS_REC	db_rec;

	dberrbuf[0] = '\0';
	if(d == (NBS_DB *)0 || d->prim == (DB *)0) {
		strcpy(dberrbuf,dbnotopen);
		return((NBS_REC *)0);
	}
	if(db_recbufsiz < siz) {
		if(db_recbuf != (char *)0)
			free(db_recbuf);
		if((db_recbuf = malloc(siz)) == (char *)0) {
			strcpy(dberrbuf,strerror(errno));
			return((NBS_REC *)0);
		}
		db_recbufsiz = siz;
	}
	if(fseek(d->rec,rec,SEEK_SET) != 0) {
		strcpy(dberrbuf,strerror(errno));
		return((NBS_REC *)0);
	}
	if(fread(db_recbuf,siz,1,d->rec) != 1) {
		strcpy(dberrbuf,strerror(errno));
		return((NBS_REC *)0);
	}
	/* align on some architectures */
	memcpy(&db_rec,db_recbuf,sizeof(db_rec));

	/* check magic */
	if(rec != db_rec.off) {
		strcpy(dberrbuf,"record magic number bad; database corrupt");
		return((NBS_REC *)0);
	}

	/* set pointers to record and comment */
	db_rec.rec = db_recbuf + sizeof(db_rec);
	db_rec.cmt = db_recbuf + sizeof(db_rec) + db_rec.rsiz;
	return(&db_rec);
}



/* put a record to the database. */
int
nbs_dbputrec(NBS_DB *d,NBS_REC *r,int siz)
{
	dberrbuf[0] = '\0';
	if(d == (NBS_DB *)0 || d->prim == (DB *)0) {
		strcpy(dberrbuf,dbnotopen);
		return(1);
	}
	if(fseek(d->rec,r->off,SEEK_SET) != 0)
		goto bail;
	/* write the struct */
	if(fwrite(r,sizeof(NBS_REC),1,d->rec) != 1)
		goto bail;
	if(r->rsiz > 0 && fwrite(r->rec,r->rsiz,1,d->rec) != 1)
		goto bail;
	if(r->csiz > 0 && fwrite(r->cmt,r->csiz,1,d->rec) != 1)
		goto bail;
	return(0);

bail:
	strcpy(dberrbuf,strerror(errno));
	return(1);
}



/* insert a record into the primary key tree */
int
nbs_dbinsprim(NBS_DB *d,char *k,int l,off_t off,int siz)
{
	NBS_PRI	b;
	int	x;

	dberrbuf[0] = '\0';
	if(d == (NBS_DB *)0 || d->prim == (DB *)0) {
		strcpy(dberrbuf,dbnotopen);
		return(1);
	}
	b.off = off;
	b.siz = siz;
	memset(&key,0,sizeof(DBT));
	memset(&data,0,sizeof(DBT));
	key.data = k;
	key.size = l - 1; /* no need to store nulls in btree */
	data.data = &b;
	data.size = sizeof(b);
	x = d->prim->put(d->prim,NULL,&key,&data,0);
        if(x != 0) {
                strncpy(dberrbuf,db_strerror(x),BUFSIZ - 2);
                return(1);
        }
	return(0);
}



/* delete a count record and insert one in the count tree */
int
nbs_dbupcounts(NBS_DB *d,int ct,off_t off,int siz,int mustdel)
{
	NBS_CNT	b;
	int	x;

	dberrbuf[0] = '\0';
	if(d == (NBS_DB *)0 || d->cnt == (DB *)0) {
		strcpy(dberrbuf,dbnotopen);
		return(1);
	}


	/* firstly if there is already a count, delete it .. */
	if(mustdel) {
		memset(&key,0,sizeof(DBT));
		memset(&data,0,sizeof(DBT));
		b.off = off;
		b.cnt = ct - 1; /* will be indexed by old count */
		key.data = &b;
		key.size = sizeof(b);

		x = d->cnt->del(d->cnt,NULL,&key,0);
        	if(x != 0) {
                	strncpy(dberrbuf,db_strerror(x),BUFSIZ - 2);
			strcat(dberrbuf," - updating counts");
                	return(1);
        	}
	}

	memset(&key,0,sizeof(DBT));
	memset(&data,0,sizeof(DBT));
	b.off = off;
	b.cnt = ct;
	key.data = &b;
	key.size = sizeof(b);
	data.data = &siz;
	data.size = sizeof(int);

	x = d->cnt->put(d->cnt,NULL,&key,&data,0);
        if(x != 0) {
                strncpy(dberrbuf,db_strerror(x),BUFSIZ - 2);
                return(1);
        }
	return(0);
}



/* delete an update record and insert one in the count tree */
int
nbs_dbupup(NBS_DB *d,time_t new,time_t old,off_t off,int siz,int mustdel)
{
	NBS_UPD	b;
	int	x;

	dberrbuf[0] = '\0';
	if(d == (NBS_DB *)0 || d->cnt == (DB *)0) {
		strcpy(dberrbuf,dbnotopen);
		return(1);
	}

	/* firstly if there is already a count, delete it .. */
	if(mustdel) {
		memset(&key,0,sizeof(DBT));
		memset(&data,0,sizeof(DBT));
		b.off = off;
		b.upd = old;
		key.data = &b;
		key.size = sizeof(b);

		x = d->upd->del(d->upd,NULL,&key,0);
        	if(x != 0) {
                	strncpy(dberrbuf,db_strerror(x),BUFSIZ - 2);
			strcat(dberrbuf," - updating updates");
                	return(1);
        	}
	}

	memset(&key,0,sizeof(DBT));
	memset(&data,0,sizeof(DBT));
	b.off = off;
	b.upd = new;
	key.data = &b;
	key.size = sizeof(b);
	data.data = &siz;
	data.size = sizeof(int);

	x = d->upd->put(d->upd,NULL,&key,&data,0);
        if(x != 0) {
                strncpy(dberrbuf,db_strerror(x),BUFSIZ - 2);
                return(1);
        }
	return(0);
}



/* I couldn't figure a decent way to make these generics
without spending as much code as just duplicating them */
int
nbs_dbwalkprim(NBS_DB *d,int rflg,int count,int masktim,void (*func)(char *,NBS_REC *))
{
	NBS_REC	*rp;
	NBS_PRI	b;
	DBC	*c;
	int	r;
	int	icnt = 0;
	int	flg;

	if(d == (NBS_DB *)0 || d->prim == (DB *)0) {
		strcpy(dberrbuf,dbnotopen);
		return(-1);
	}

	/* which way shall we iterate? */
	if(rflg == 0)
		flg = DB_NEXT;
	else
		flg = DB_PREV;

	dberrbuf[0] = '\0';
	r = d->prim->cursor(d->prim,NULL,&c,0);
	if(r != 0) {
		strncpy(dberrbuf,db_strerror(r),BUFSIZ - 2);
		return(1);
	}
	memset(&key,0,sizeof(DBT));
	memset(&data,0,sizeof(DBT));

	while(1) {
		r = c->c_get(c,&key,&data,flg);
		if(r != 0 && r != DB_NOTFOUND) {
			strncpy(dberrbuf,db_strerror(r),BUFSIZ - 2);
			break;
		}
		if(r == DB_NOTFOUND) {
			/* hack */
			r = 0;
			break;
		}
		/* get record and print it */
		memcpy(&b,data.data,sizeof(b));
		rp = nbs_dbgetrec(d,b.off,b.siz);
		if(rp == (NBS_REC *)0)
			return(1);

		/* apply time mask */
		if(db_timenow != (time_t)0 && masktim != -1) {
			if(db_timenow - rp->up >= masktim)
				continue;
		}
		if(count != 0) {
			if(icnt++ >= count)
				break;
		}
		(*func)(rp->rec,rp);
	}
	if(c->c_close(c)) {
		strncpy(dberrbuf,db_strerror(r),BUFSIZ - 2);
		return(1);
	}
	return(r);
}



int
nbs_dbwalkcnt(NBS_DB *d,int rflg,int count,int masktim,void (*func)(char *,NBS_REC *))
{
	NBS_REC	*rp;
	NBS_CNT	b;
	DBC	*c;
	int	r;
	int	icnt = 0;
	int	flg;
	int	siz;

	if(d == (NBS_DB *)0 || d->cnt == (DB *)0) {
		strcpy(dberrbuf,dbnotopen);
		return(-1);
	}

	if(rflg == 0)
		flg = DB_NEXT;
	else
		flg = DB_PREV;

	dberrbuf[0] = '\0';
	r = d->cnt->cursor(d->cnt,NULL,&c,0);
	if(r != 0) {
		strncpy(dberrbuf,db_strerror(r),BUFSIZ - 2);
		return(1);
	}
	memset(&key,0,sizeof(DBT));
	memset(&data,0,sizeof(DBT));

	while(1) {
		r = c->c_get(c,&key,&data,flg);
		if(r != 0 && r != DB_NOTFOUND) {
			strncpy(dberrbuf,db_strerror(r),BUFSIZ - 2);
			break;
		}
		if(r == DB_NOTFOUND) {
			/* hack */
			r = 0;
			break;
		}
		memcpy(&b,key.data,sizeof(b));
		memcpy(&siz,data.data,sizeof(siz));
		rp = nbs_dbgetrec(d,b.off,siz);
		if(rp == (NBS_REC *)0)
			return(1);

		if(db_timenow != (time_t)0 && masktim != -1) {
			if(db_timenow - rp->up >= masktim)
				continue;
		}
		if(count != 0) {
			if(icnt++ >= count)
				break;
		}
		(*func)(rp->rec,rp);
	}
	if(c->c_close(c)) {
		strncpy(dberrbuf,db_strerror(r),BUFSIZ - 2);
		return(1);
	}
	return(r);
}



int
nbs_dbwalkupd(NBS_DB *d,int rflg,int count,int masktim,void (*func)(char *,NBS_REC *))
{
	NBS_REC	*rp;
	NBS_UPD	b;
	DBC	*c;
	int	r;
	int	icnt = 0;
	int	flg;
	int	siz;

	if(d == (NBS_DB *)0 || d->upd == (DB *)0) {
		strcpy(dberrbuf,dbnotopen);
		return(-1);
	}

	if(rflg == 0)
		flg = DB_NEXT;
	else
		flg = DB_PREV;

	dberrbuf[0] = '\0';
	r = d->upd->cursor(d->upd,NULL,&c,0);
	if(r != 0) {
		strncpy(dberrbuf,db_strerror(r),BUFSIZ - 2);
		return(1);
	}
	memset(&key,0,sizeof(DBT));
	memset(&data,0,sizeof(DBT));

	while(1) {
		r = c->c_get(c,&key,&data,flg);
		if(r != 0 && r != DB_NOTFOUND) {
			strncpy(dberrbuf,db_strerror(r),BUFSIZ - 2);
			break;
		}
		if(r == DB_NOTFOUND) {
			/* hack */
			r = 0;
			break;
		}
		memcpy(&b,key.data,sizeof(b));
		memcpy(&siz,data.data,sizeof(siz));

		/* we have the update times in the db already */
		if(db_timenow != (time_t)0 && masktim != -1) {
			if(db_timenow - b.upd >= masktim)
				continue;
		}
		rp = nbs_dbgetrec(d,b.off,siz);
		if(rp == (NBS_REC *)0)
			return(1);

		if(count != 0) {
			if(icnt++ >= count)
				break;
		}
		(*func)(rp->rec,rp);
	}
	if(c->c_close(c)) {
		strncpy(dberrbuf,db_strerror(r),BUFSIZ - 2);
		return(1);
	}
	return(r);
}



static	int
shuffle(char *from,char *to,char *suf,int bkup)
{
	char	tbuf[BUFSIZ];
	char	fbuf[BUFSIZ];

	/* this is ugly but need not be pretty or fast */
	sprintf(fbuf,"%s.%s",from,suf);
	sprintf(tbuf,"%s.%s",to,suf);
	if(bkup) {
		char	bbuf[BUFSIZ];

		sprintf(bbuf,"%s.%s.BAK",to,suf);
		(void)unlink(bbuf);
		if(rename(fbuf,bbuf))
			return(1);
	} else {
		(void)unlink(tbuf);
		if(rename(fbuf,tbuf))
			return(1);
	}
	return(0);
}



/* move a set of db files */
int
nbs_dbrename(char *from,char *to,int flg,int bkup)
{
	if(shuffle(from,to,"rec",bkup))
		return(1);
	if(shuffle(from,to,"idx",bkup))
		return(1);
	if(shuffle(from,to,"upd",bkup))
		return(1);
	if(shuffle(from,to,"cnt",bkup))
		return(1);
	return(0);
}
