#include	<stdio.h>
#include	<ctype.h>


/*
	Marcus J. Ranum, 2004   - All rights reserved
	$Id: gtemp.c,v 1.1 2004/09/03 01:16:19 mjr Exp $

	This software may be used and redistributed free of charge,
	but may not be incorporated into a commercial product or
	offering without the author's permission. Please see
	the included "LICENSE.pdf" for the software's governing
	license and terms.

	NBS - "Never Before Seen" anomaly detector
	this module generates a simple "structural template"
	from a string. This is used for the weird-o structural
	mode in nbs. better algorithms can be used if this
	proves useful.
*/

static	char	*out;	/* output pointer */
static	int	outc;	/* ouput size */
static	int	smax;	/* max output size */
static	char	*i;	/* input pointer */


static	int
p_str()
{
	while(1) {
		if(isalnum((int)*i)) {
			i++;
			continue;
		}

		if(outc > smax - 2) {
			*out = '\0';
			return(-1);
		}
		outc++;
		*out++ = '%';
		outc++;
		*out++ = 's';
		return(0);
	}
}



static	int
p_num()
{
	char	*sav = i;	/* saved location in input */

	while(1) {
		if(isdigit((int)*i)) {
			i++;
			continue;
		}

		/* if we got an alpha
		it's time to rewind and
		jump over to treat this
		as a mixed alpha/num
		string */
		if(isalpha((int)*i)) {
			i = sav;
			return(p_str());
		}

		/* must be a special
		character. we are done. */
		if(outc > smax - 2) {
			*out = '\0';
			return(-1);
		}
		outc++;
		*out++ = '%';
		outc++;
		*out++ = 'd';
		return(0);
	}
}



static	int
q_str()
{
	int	quot = *i++;

	while(1) {
		/* SHOOT! ran off end of string
		attempt recovery */
		if(*i == '\n' || *i == '\0')
			return(1);

		if(*i == quot) {
			i++;
			break;
		}

		/* attempt to handle \" quotes */
		if(*i == '\\' && *(i + 1) == quot)
			i++;

		i++;
	}
	if(outc > smax - 2) {
		*out = '\0';
		return(-1);
	}
	outc++;
	*out++ = '%';
	outc++;
	*out++ = 'q';
	return(0);
}



int
gtemplate(char *in,char *tp,int max,int *l)
{
	outc = 0;
	out = tp;
	smax = max;
	i = in;

	while(1) {
		if(*i == '\0') {
			*out = '\0';
			*l = outc + 1;
			return(0);
		}

		if(*i == '\n' && *(i + 1) != '\0') {
			i++;
			if(outc > smax - 2) {
				*out = '\0';
				return(-1);
			}
			outc++;
			*out++ = '%';
			outc++;
			*out++ = '$';
			continue;
		}

		/* if we have whitespace
		output a whitespace delim */
		if(isspace((int)*i)) {
			while(isspace((int)*i))
				i++;
			if(outc > smax - 1) {
				*out = '\0';
				return(-1);
			}
			outc++;
			*out++ = ' ';
			continue;
		}

		if(isdigit((int)*i)) {
			p_num();
			continue;
		}

		/* quoted string? - ugly
		save state and if we fail
		we fall through to plain str */
		if(*i == '"' || *i == '\'') {
			char	*sav = i;
			switch(q_str()) {
			case -1:
				return(-1);
			case 0:
				continue;
			case 1:
				/* FALL THRU */
				i = sav;
				break;
			}
		}

		/* collect a string with a
		kludge for pathnames */
		if(isalpha((int)*i)) {
			p_str();
			continue;
		}

		/* %? */
		if(*i == '%') {
			if(outc > smax - 2) {
				*out = '\0';
				return(-1);
			}
			outc += 2;
			*out++ = '%';
			*out++ = *i++;
			continue;
		}

		/* it's something else: try it as a literal */
		if(outc > smax - 1) {
			*out = '\0';
			return(-1);
		}
		outc++;
		*out++ = *i++;
	}
}
