#include	<sys/file.h>
#include	<sys/types.h>
#include	<time.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<string.h>
#include	<stdio.h>

#include	<db.h>

#include	"nbs.h"

extern	int	errno;
extern	char	*optarg;

/*
	Marcus J. Ranum, 2004   - All rights reserved
	$Id: nbs.c,v 1.1 2004/09/03 01:16:19 mjr Exp $

	This software may be used and redistributed free of charge,
	but may not be incorporated into a commercial product or
	offering without the author's permission. Please see
	the included "LICENSE.pdf" for the software's governing
	license and terms.

	NBS - "Never Before Seen" anomaly detector
*/


static	void
usage()
{
	fprintf(stderr,"usage: nbs [options]\n");
	fprintf(stderr,"\t[-d dbfile] use database 'dbfile'\n");
	fprintf(stderr,"\t[-i infile] read input from 'infile'\n");
	fprintf(stderr,"\t[-o outfile] send output to 'outfile'\n");
	fprintf(stderr,"\t[-a outfile] append output to 'outfile'\n");
	fprintf(stderr,"\t[-e errfile] append error messages to 'errfile'\n");
	fprintf(stderr,"\t[-D] debug\n");
	fprintf(stderr,"\t[-E] exit with number of NBS as return value (-1 = fail)\n");
	fprintf(stderr,"\t[-s] summarize upon completion\n");
	exit(-1);
}

time_t	atthetonethetimewillbe;

static	int	nbsupdate(NBS_DB *,char *,int,char *,int);
static	int	dflg = 0;		/* debug flag */
static	int	eflg = 0;		/* exit with NBS flag */
static	char	*dbfilename = NBS_DFLTFILE;
static	char	*infilename = "standard input";
static	int	nb_seen = 0;
static	FILE	*erf;
static	FILE	*ouf;


int
main(int ac, char *av[])
{
	int	x;
	int	x2;
	NBS_DB	*d;
	FILE	*inf;
	char	buf[BUFSIZ];
	char	sbuf[BUFSIZ];
	int	lno = 0;
	int	sflg = 0;

	inf = stdin;
	ouf = stdout;
	erf = stderr;
	time(&atthetonethetimewillbe);
	while ((x = getopt(ac, av, "DEsa:d:e:i:o:")) != -1)
		switch(x) {
		case 'd':
			dbfilename = optarg;
			break;
		case 'i':
			infilename = optarg;
			if((inf = fopen(optarg,"r")) == (FILE *)0) {
				perror(optarg);
				exit(-1);
			}
			break;
		case 'o':
			if((ouf = fopen(optarg,"w")) == (FILE *)0) {
				perror(optarg);
				exit(-1);
			}
			break;
		case 'a':
			if((ouf = fopen(optarg,"a")) == (FILE *)0) {
				perror(optarg);
				exit(-1);
			}
			break;
		case 'e':
			if((erf = fopen(optarg,"w")) == (FILE *)0) {
				perror(optarg);
				exit(-1);
			}
		case 'D':
			dflg = 1;
			fprintf(stderr,"Debugging turned on\n");
			break;
		case 's':
			sflg = 1;
			break;
		case 'E':
			eflg = 1;
			break;
		default:
			fprintf(stderr,"bad argument: -%c\n",x);
			usage();
		}


	if((d = nbs_dbopen(dbfilename)) == (NBS_DB *)0) {
		fprintf(stderr,"failed to open database: %s\n",nbs_dberror());
		exit(-1);
	}

	while(1) {
		if(fgets(buf,sizeof(buf),inf) == (char *)0)
			break;
		buf[sizeof(buf) - 2] = '\0';

		x = strlen(buf);
		lno++;
		d->tlin++;

		/* skip it if not newline terminated */
		if(buf[x - 1] != '\n') {
			fprintf(stderr,"%s line %d: no newline - possibly too long\n",infilename,lno);
			continue;
		}
		buf[x - 1] = '\0';

		/* in structure mode, generate and store IT as the NBS... */
		if((d->flg & NBS_DB_FLG_STRUCTUR) != 0) {
			if(gtemplate(buf,sbuf,sizeof(sbuf),&x2)) {
				fprintf(erf,"template generation failed - %s line %d\n",infilename,lno);
				break;
			}
			if(nbsupdate(d,sbuf,x2,buf,x))
				break;
			continue;
		}

		if(nbsupdate(d,buf,x,(char *)0,0))
			break;
	}

	/* some updates! - saved when we close */
	if(nb_seen)
		d->lupd = atthetonethetimewillbe;
	d->lopd = atthetonethetimewillbe;
	if(nbs_dbclose(d)) {
		fprintf(erf,"failed to close database: %s\n",nbs_dberror());
		exit(-1);
	}

	if(sflg) {
		time_t	dtim;
		float	prct;

		time(&dtim);
		fprintf(erf,"NBS completed run (%d sec)\n",dtim - atthetonethetimewillbe);
		fprintf(erf,"%d never before seen entries\n",nb_seen);
		prct = ((float)nb_seen / (float)lno) * 100;
		fprintf(erf,"%d lines of input processed (%.2f%% NBS)\n",lno,prct);
		fprintf(erf,"%d total entries in database\n",d->ecnt);
	}

	if(eflg)
		exit(nb_seen);

	/* TODO - database close */
	exit(0);
}



/* most of the work is done here */
static	int
nbsupdate(NBS_DB *d,char *l,int len,char *l2, int len2)
{
	off_t	rec;
	int	siz;
	NBS_REC	*rp;
	NBS_REC	b;

	/* since we use offsets for records they will be positive */
	if((rec = nbs_dbck(d,l,len,&siz)) > 0) {
		time_t	lastup;

		if(d->flg & NBS_DB_FLG_NOREC)	/* don't update anything */
			return(0);
		
		if((rp = nbs_dbgetrec(d,rec,siz)) == (NBS_REC *)0) {
			fprintf(erf,"record %d get failed! %s\n",(int)rec,nbs_dberror());
			return(1);
		}

		/* update counts */
		rp->cnt += 1;
		/* update update */
		lastup = rp->up;
		rp->up = atthetonethetimewillbe;

		/* put it back */
		if(nbs_dbputrec(d,rp,siz)) {
			fprintf(erf,"record %d put failed! %s\n",(int)rec,nbs_dberror());
			return(1);
		}
		if(dflg)
			fprintf(stderr,"updated %d count=%d size=%d\n",(int)rec,rp->cnt,siz);

		/* no need to update primary index */
		/* update counts and upd index */
		if((d->flg & NBS_DB_FLG_NOCOUNT) == 0 && nbs_dbupcounts(d,rp->cnt,rp->off,siz,1)) {
			fprintf(erf,"record %d put failed! %s\n",(int)rec,nbs_dberror());
			return(1);
		}
		if((d->flg & NBS_DB_FLG_NOUPTIM) == 0 && nbs_dbupup(d,rp->up,lastup,rp->off,siz,1)) {
			fprintf(erf,"record %d put failed! %s\n",(int)rec,nbs_dberror());
			return(1);
		}
		return(0);
	}

	/* error */
	if(rec == -1) {
		fprintf(erf,"primary key check failed! %s\n",nbs_dberror());
		return(1);
	}

	/* another entry in the total count for this db */
	d->ecnt++;

	/* if we are not to keep records, just make a dummy primary */
	if(d->flg & NBS_DB_FLG_NOREC) {
		rec = b.off = d->ecnt;
		b.cnt = -1;
		siz = 0;
	} else {
		rec = nbs_dbnewrec(d,len,len2,&siz);
		b.off = rec;
		b.frst = b.up = atthetonethetimewillbe;
		b.cnt = 1;
		b.rsiz = len;
		b.csiz = len2;
		b.rec = l;
		b.cmt = l2;
		if(nbs_dbputrec(d,&b,siz)) {
			fprintf(erf,"record %d put failed! %s\n",(int)rec,nbs_dberror());
			return(1);
		}
	}

	/* need to add to primary index */
	if(nbs_dbinsprim(d,l,len,b.off,siz)) {
		fprintf(erf,"primary index put %d failed! %s\n",(int)rec,nbs_dberror());
		return(1);
	}
	if((d->flg & NBS_DB_FLG_NOCOUNT) == 0 && nbs_dbupcounts(d,b.cnt,b.off,siz,0)) {
		fprintf(erf,"record %d put failed! %s\n",(int)rec,nbs_dberror());
		return(1);
	}
	if((d->flg & NBS_DB_FLG_NOUPTIM) == 0 &&
		nbs_dbupup(d,atthetonethetimewillbe,atthetonethetimewillbe,b.off,siz,0)) {
		fprintf(erf,"record %d put failed! %s\n",(int)rec,nbs_dberror());
		return(1);
	}

	if(dflg)
		fprintf(stderr,"new entry %d count=%d size=%d\n",(int)rec,b.cnt,siz);


	/* seen a new one; output a record */
	fprintf(ouf,"%s\n",l);
	nb_seen++;
	return(0);
}
