/*
	Marcus J. Ranum, 2004   - All rights reserved
	$Id: nbs.h,v 1.1 2004/09/03 01:16:19 mjr Exp $

	This software may be used and redistributed free of charge,
	but may not be incorporated into a commercial product or
	offering without the author's permission. Please see
	the included "LICENSE.pdf" for the software's governing
	license and terms.

	NBS - "Never Before Seen" anomaly detector
	header and forward decls
*/

#define	NBS_DFLTFILE	"neverseen"

#define	NBS_UPDSCORE	24

/* if you turn this off the DB will not be
locked. which means you can read a Btree
that is partly updated. the results are
not pretty. don't do it. */
#define	NBS_DEFAULTLOCK


typedef	struct	{
	DB	*prim;	/* primary index */
	DB	*cnt;	/* count index */
	DB	*upd;	/* update index */
	FILE	*rec;	/* record file */

	/* mirrored from the header */
	int	flg;
	int	ecnt;	/* entries in it */
	int	tlin;	/* total lines of input consumed */
	time_t	lupd;	/* last updated */
	time_t	lopd;	/* last opened */
	time_t	cred;	/* created */
} NBS_DB;
#define	NBS_DB_FLG_NOUPTIM	001
#define	NBS_DB_FLG_NOCOUNT	002
#define	NBS_DB_FLG_NOREC	004
#define	NBS_DB_FLG_STRUCTUR	010


/* a record containing complete details about an NBS */
typedef	struct	{
	off_t	off;			/* sanity check */
	time_t	up;			/* last update */
	time_t	frst;			/* first installed */
	int	cnt;			/* count */
	int	rsiz;			/* record length */
	int	csiz;			/* record length */
	char	*rec;			/* pointer to record */
	char	*cmt;			/* pointer to comment */
} NBS_REC;


/* struct stored in the primary btree */
typedef	struct	{
	off_t	off;
	int	siz;
} NBS_PRI;


/* struct stored in the count btree */
typedef	struct	{
	int	cnt;			/* count */
	off_t	off;
} NBS_CNT;


/* struct stored in the update times btree */
typedef	struct	{
	time_t	upd;			/* last update time */
	off_t	off;
} NBS_UPD;


/* header for the record file */
typedef	struct	{
	char	warn[128];	/* misc text */
	int	flg;
	int	ecnt;		/* entries in it */
	int	tlin;		/* total lines of input consumed */
	time_t	lupd;		/* last updated */
	time_t	lopd;		/* last opened */
	time_t	cred;		/* created */
} NBS_HDR;


/* forward decls - db.c */
extern	NBS_DB	*nbs_dbcreat(char *,int);
extern	NBS_DB	*nbs_dbopen(char *);
extern	NBS_REC	*nbs_dbgetrec(NBS_DB *,off_t,int);
extern	off_t	nbs_dbck(NBS_DB *,char *,int,int *);
extern	off_t	nbs_dbnewrec(NBS_DB *,int,int,int *);
extern	char	*nbs_dberror();
extern	int	nbs_dbclose(NBS_DB *);
extern	int	nbs_dbputrec(NBS_DB *,NBS_REC *,int);
extern	int	nbs_dbinsprim(NBS_DB *,char *,int,off_t,int);
extern	int	nbs_dbupcounts(NBS_DB *,int,off_t,int,int);
extern	int	nbs_dbupup(NBS_DB *,time_t,time_t,off_t,int,int);
extern	int	nbs_dbwalkprim(NBS_DB *,int,int,int,void (*func)(char *,NBS_REC *));
extern	int	nbs_dbwalkcnt(NBS_DB *,int,int,int,void (*func)(char *,NBS_REC *));
extern	int	nbs_dbwalkupd(NBS_DB *,int,int,int,void (*func)(char *,NBS_REC *));
extern	int	nbs_dbrename(char *,char *,int,int);

/* gtemp.c */
extern	int	gtemplate(char *,char *,int,int *);
