.TH NBS 1 "updated 04 July 2004"
.SH NAME
NBS \- Never Before Seen anomaly detector suite
.SH SYNOPSIS
.B nbs
[ options ]
.br
.B nbsmk
[ options ]
.br
.B nbsdump
[ options ]
.br
.B nbspreen
[ options ]
.SH DESCRIPTION
.I  NBS
implements a fast look-up database for arbitrary strings,
and outputs strings it has never seen before. This capability
is useful for system log analysis, network event detection
or monitoring, or other security/system administration
tasks.
.PP
There are few built-in limits to NBS; the database's limits
are those of the underlying BSD-DB B-tree index, and the
filesystem. Theoretically databases of millions of records and
terabytes of information should function correctly and be
fairly quickly searchable. The underlying records are stored
in a flat file with hard offsets encoded in the B-trees; the
files are thus not machine portable, and are optimized for
check-time performance and append. Deleting records is not
supported except as part of a batch file-preening operation
using
.I nbspreen
to expire records based on last update time.
.PP
A typical cycle of use for the
.I NBS
suite is to create an
.I NBS
database with
.I nbsmk
then to use
.I nbs
to check status against and update its contents.
The
.I nbsdump
program might be used in a batch process to dump the
most (or least-) frequently seen items, for reporting
purposes. Lastly, the
.I nbspreen
utility might be used to periodically "expire" old
entries so that they will show up again as "never before
seen" if they occur extremely intermittently (e.g.: annually
or monthly).
.PP
When an
.I NBS
database is created, the database contains the settings
that control how it is to be updated. Thus, it is important
to correctly configure the database to suit your purpose.
In most cases, the defaults will provide the performance
and flexibility you need. When an
.I NBS
database is created, the utility creates several files,
e.g.:
.br
.I
neverseen.cnt
\- the counts index (allows rapid sorting/retrieval of
"bottom/top N" data)
.br
.I
neverseen.idx
\- the master index (allows rapid neverseen checking and
sorted retrieval of data keys)
.br
.I
neverseen.rec
\- the data records and their update information
.br
.I
neverseen.upd
\- the update time index (allows rapid sorting/retrieval
of "least/most recently" updated data)
.PP
In extreme performance situations, it is possible to ask
.I NBS
to not update some of the indexes, or even the primary
record database. It is possible to run
.I NBS
in a mode so that the only index that is kept is the primary
B-tree; though this is seldom necessary. Depending on the
indexes that are enabled or disabled, some of the files in
a given database may be absent, or may remain small and
never grow. Users should only worry about
leaving indexes out if NBS appears to be a
performance or disk space hog once tested operationally.
User experience shows this is seldom an issue.
.PP
Typical applications of the
.I NBS
suite are to track DHCP assignments, inter-host connectivity, URLs
seen in web server logs, sender and recipient addresses of Emails,
attachment names, IDS alerts seen, etc.
.PP
.SH "NBSMK Options:"
.PP
\fB\-d database\fP 
specify the root name for the database to create. The default
is \fBneverseen\fP. Databases may be fully-pathed.
.PP
\fB\-K\fP
create the database to not maintain the count B-tree (".cnt") file.
.PP
\fB\-U\fP
create the database to not maintain the update B-tree (".upd") file.
.PP
\fB\-0\fP (zero)
create the database to not maintain the record file. In this case
the ".rec" file is created so it can contain set-up information and
be used to synchronize access via locking, but will not be updated
with records.
.PP
\fB\-S\fP
create the database to store records in
.I structural analysis mode
which is an experimental technique in which
.I
NBS
takes its input and converts it into an abstract printf-like
string that discards the numbers and strings while replacing
them with template specifiers such as "%d". Thus, the string
\fB"21:24:35.715495 216.193.194.13.ftp-data"\fP
would get converted to:
\fB"%d:%d:%d.%d %d.%d.%d.%d.%s-%s\fP
and only new templates would be logged when discovered. This
mode may be useful for detecting corruptions in highly
structured data.
.PP
\fB\-D\fP
toggles some diagnostic/debug output.
.PP
.SH "NBS Options:"
.PP
\fB\-d database\fP 
specify the root name for the database to use. The database
must already have been created with
.I nbsmk.
.PP
\fB\-i infile\fP 
specify the file from which to read input. Default is standard
input. Inputs are considered to be line-oriented non-binary data.
.PP
\fB\-o outfile\fP 
specify the file to which to write output. Default is standard
output. Standard/IO buffering is performed on output; while the
output is line-oriented, no guarantees are made as to when
buffers are flushed. This option truncates the output file.
.PP
\fB\-a outfile\fP 
is the same as \fB-o\fP except that the output file is appended
to, rather than truncated.
.PP
\fB\-e errfile\fP 
specify the file to which error messages should be written.
The error file is truncated.
.PP
\fB\-E\fP
exit with the number of never before seen lines as the
exit status (or -1 in the event of an error). Normal
exit status is 0 (success) or -1 (failure).
.PP
\fB\-s\fP
Output summary statistics about the database and percentage
if the input lines that were never before seen.
.PP
\fB\-D\fP
toggles some diagnostic/debug output.
.PP
.SH "NBSDUMP Options:"
.PP
\fB\-d database\fP 
specify the root name for the database to use. The database
must already have been created with
.I nbsmk.
.PP
\fB\-k key\fP
specify a single key to look for and dump if it exists. This
would generally be combined with the -V option. For example,
if you wanted to know how many times "index.html" was
accessed you could use
\fBnbsdump \-V \-k index.html\fP
.PP
\fB\-c #\fP
specify the number of entries to dump. This is generally
used to get top/bottom counts (e.g.: "top ten list") and
is often combined with the \fB-R\fP flag to control
ordering. Depending on the particular value that is
requested (default is the key string) the ordering will
be either the key string, the count, or the update time.
.PP
\fB\-K\fP
order output by the sorting sequence of the keys in
the database (alphabetic or reverse-alphabetic). This
flag is actually a no-op, since \fB-K\fP is the default.
.PP
\fB\-C\fP
order output by the number of times that the key has
been seen (count). This builds a sorted frequency
chart for the key.
.PP
\fB\-U\fP
order output by the update times for the record. This
builds a "most/least recently updated" list.
.PP
\fB\-R\fP
reverse the sort order. Default sort order is small to
large. If you want the top 10 items from the database
you would use
\fBnbsdump -c 10 -R\fB
.PP
\fB\-T dd:hh:mm:ss\fP
apply a time mask to the dump; disregard records older
(in seconds) than the specified time. Values can be
omitted from the right hand side; i.e:
\fB\-T 30\fP would disregard records older than 30 days.
\fBnbsdump -c 10 -R -T 60:0:0:1\fP would print the first 10
records, in reverse order, that had been updated within
the last sixty days and one second. Bear in mind that this
type of query will result in a "search until satisfied"
traversal of the index so performance may suffer.
.PP
\fB\-V\fP
turns on more verbose output. With this flag selected
the complete contents of the database record are
printed out (whatever is available).
.PP
\fB\-M\fP
turns on minimalist output. With this flag selected,
.I nbsdump
tries to be as succinct as possible (usually just
printing a key or count).
.PP
\fB\-X\fP
outputs in pseudo-XML-oid format for those who
insist on XML-ish markup.
.PP
\fB\-Y\fP
outputs in year/date output where the date is the
leading value.
.PP
\fB\-x\fP
outputs in "technical" format. Technical format
is the anti-XML; it's \fBattribute=value\fP pairs
suitable for extracting with shell scripts.
.PP
\fB\-s\fP
append statistics about the database's contents and
state when completed.
.PP
\fB\-S\fP
append statistics about the database \fBonly\fP. No
records are printed.
.PP
\fB\-D\fP
toggles debugging output.
.PP
.SH "NBSPREEN Options:"
.PP
\fB\-T d:h:m\fP
\fImandatory\fP
option specifies the cutoff for how old records should
be in order to be purged. If fields are omitted, they
are parsed left-to-right, I.e:
\fB\-T 365\fP means "discard anything that has not
been updated in a year" whereas
\fB\-T 0:24\fP means "discard if older than 24 hours."
.PP
\fB\-d database\fP
specifies the database to use. The database must
already exist.
.PP
\fB\-b\fP
specifies that the old versions of the database
should be renamed +".BAK" instead of deleted in
the update process.
.PP
\fB\-v\fP
turns on verbose output.
.PP
\fB\-D\fP
turns on diagnostic/debug output.
.PP
The
.I nbspreen
utility creates a new database and copies all the records
from the original into it, if they are newer than a specified
cut-off time. Once the operation is complete, the new database
replaces the old. By design, the
.I NBS
system should not require frequent database rebuilds. I.e.:
.I nbspreen
is not a performance-enhancing tool; it is a tool for
allowing you to decide a new value of "never before."
This might come in handy if, say, you want to track
DHCP leases but wish to be notified of repeat leases
that occur very rarely - say every 3 months or so.
In that case, you would rebuild the database with
.I nbspreen
every day, week, or month, discarding everything
older than 3 months with \fB\-T 90\fP. Be aware
that rebuilding the database requires reading
and rewriting the entire database as a new file;
make sure there is enough disk space and that
the I/O load will not be burdensome.
.SH "EXAMPLES"
.PP
.I NBS
can be used for a wide variety of applications
but is primarily aimed toward anomaly detection.
For a simple example, let's suppose we want to detect
new access requests against web pages in our
Apache-based web server.
.PP
Log messages on this system are in the form:
.br
\fI10.10.10.10 - - [02/Dec/2002:20:39:56 -0500] "GET / HTTP/1.1" 200 2187\fP
so we write a simple perl script to pull out the URLs:
.br
.nf
\fI
#!/usr/bin/perl
while (<>) {
        if(/"GET.*HTTP/) {
                @array = split(/ /);
                print("$array[6]\\n");
        }
}
\fP
.fi
.PP
This perl script dumps out URLs when run as follows:
.br
retail /var/www/logs/access_log | geturl.pl
.PP
We then create an
.I NBS
database to store the URLs:
.br
\fInbsmk\fP
.br
and populate it from a cron job as follows:
.br
\fIretail /var/www/logs/access_log | geturl.pl | nbs -o new\fP
.PP
Any new records that are introduced into the database
will be stored in the file "new" which can be Emailed
to an administrator.
.PP
If we subsequently wish to have a periodic Emailing of
the top-ten URLs in the database, we might use:
.br
\fInbsdump -c 10 -r | mail -s "top ten URLs" mjr\fP
.SH "SEE ALSO"
retail
.SH BUGS
.PP
There are line-length limitations inherent in nbs'
main input routine. Lines longer than BUFSIZ - 1
will be truncated and may get treated as multiple
lines of input. While nbs' database should be
able to handle large amounts of information, you
should be conscious of the size of the records;
larger records in the main index will impact
performance and disk use.
.PP
It would be nice if there was a way for NBS to
retain an entire line of input but only perform
NBS parsing on a subset of it. Unfortunately doing
that would require an internal parser; it seems
better to keep parsing external.
.br
