#include	<sys/file.h>
#include	<sys/types.h>
#include	<time.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<string.h>
#include	<stdio.h>

#include	<db.h>

#include	"nbs.h"

extern	int	errno;

/*
	Marcus J. Ranum, 2004   - All rights reserved
	$Id: nbsmk.c,v 1.1 2004/09/03 01:16:19 mjr Exp $

	This software may be used and redistributed free of charge,
	but may not be incorporated into a commercial product or
	offering without the author's permission. Please see
	the included "LICENSE.pdf" for the software's governing
	license and terms.

	NBS - "Never Before Seen" anomaly detector database initializer
*/


static	void
usage()
{
	fprintf(stderr,"usage: mknbs [options]\n");
	fprintf(stderr,"\t[-d database] specify database\n");
	fprintf(stderr,"\t[-K] do not maintain count index\n");
	fprintf(stderr,"\t[-U] do not maintain update time(s)\n");
	fprintf(stderr,"\t[-0] do not maintain records at all\n");
	fprintf(stderr,"\t[-S] maintain structure mode (experimental)\n");
	fprintf(stderr,"\t[-D] debug\n");
	exit(1);
}


static	int	dflg = 0;		/* debug flag */
static	char	*dbfilename = NBS_DFLTFILE;


int
main(int ac, char *av[])
{
	int	x;
	NBS_DB	*d;
	int	flg = 0;

	while ((x = getopt(ac, av, "SDd:UK0O")) != -1)
		switch(x) {
		case 'd':
			dbfilename = optarg;
			break;
		case 'D':
			dflg = 1;
			fprintf(stderr,"Debugging turned on\n");
			break;
		case 'K':
			flg |= NBS_DB_FLG_NOCOUNT;
			break;
		case 'U':
			flg |= NBS_DB_FLG_NOUPTIM;
			break;
		case '0':
		case 'O':
			flg |= (NBS_DB_FLG_NOCOUNT|NBS_DB_FLG_NOUPTIM|NBS_DB_FLG_NOREC);
			break;
		case 'S':
			flg |= NBS_DB_FLG_STRUCTUR;
			break;
		default:
			fprintf(stderr,"bad argument: -%c\n",x);
			usage();
		}

	if((flg & NBS_DB_FLG_NOREC) && (flg & NBS_DB_FLG_STRUCTUR)) {
		fprintf(stderr,"-S flag and -0 flag are mutually exclusive\n");
		exit(1);
	}

	/* don't need no stinkin' "e" */
	if((d = nbs_dbcreat(dbfilename,flg)) == (NBS_DB *)0) {
		fprintf(stderr,"failed to create database: %s\n",nbs_dberror());
		exit(1);
	}
	exit(0);
}
